import { DynamoDBDocumentClient, QueryCommandInput, ScanCommandInput } from '@aws-sdk/lib-dynamodb';
import { DynamoDBServiceException } from '@aws-sdk/client-dynamodb';
export interface SimpleDocumentType {
    id: string;
    createdAt: string;
    updatedAt: string;
}
type FindAllByQuery = {
    lastEvaluatedId?: string;
    scanCommandInput: Omit<ScanCommandInput, 'TableName'>;
};
type PutRequest = {
    PutRequest: {
        Item: SimpleDocumentType;
    };
};
type DeleteRequest = {
    DeleteRequest: {
        Key: {
            id: string;
        };
    };
};
type WriteRequest = PutRequest | DeleteRequest;
export type BackOffProvider = (retries: number) => Promise<void>;
export interface SimpleDynamoRepositoryOptions {
    backOffProvider?: BackOffProvider;
    maxRetries?: number;
    batchSize?: number;
}
interface InternalBatchExecutionResult<RequestType extends WriteRequest> {
    unprocessed: RequestType[];
}
/**
 * The result of a batch execution.
 * Includes document ids of unprocessed items.
 */
export interface BatchExecutionResult {
    unprocessed: string[];
}
/**
 * A simple repository for DynamoDB
 * @template DocumentType the type of the document
 * example:
 * ```ts
 * type User {
 *  id: string
 *  name: string
 * }
 * class UserRepository extends SimpleDynamoRepository<User> {}
 * const userRepository = new UserRepository(new DynamoDBClient({ region: 'eu-central-1' }), 'users')
 * ```
 */
export declare abstract class SimpleDynamoRepository<DocumentType extends SimpleDocumentType> {
    protected readonly client: DynamoDBDocumentClient;
    protected readonly tableName: string;
    protected backOffProvider: BackOffProvider;
    protected maxRetries: number;
    protected batchSize: number;
    constructor(client: DynamoDBDocumentClient, tableName: string, options?: SimpleDynamoRepositoryOptions);
    /**
     * Find a document by id
     * @param id
     * @throws DocumentNotFoundError if the document does not exist
     * @returns the document
     */
    findOne(id: string): Promise<DocumentType>;
    /**
     * Find a document by the global secondary index
     * @param key the name of the secondary index
     * @param value the value of the secondary index
     * @returns all documents with the given secondary index
     */
    findBy(secondaryKey: string, value: string): Promise<DocumentType[]>;
    /**
     * Find the first document by the global secondary index
     * @param key the name of the secondary index
     * @param value the value of the secondary index
     * @returns the first document with the given secondary index
     */
    findOneBy(secondaryKey: string, value: string): Promise<DocumentType>;
    /**
     * Save a document
     * @param doc
     * @returns the saved document
     */
    save(doc: DocumentType): Promise<DocumentType>;
    /**
     * Save a list of documents in batches.
     * It will send batches of 25 items synchronously.
     * @param documents list of documents that will be saved in batches
     * @returns {BatchExecutionResult}
     */
    batchSave(documents: DocumentType[]): Promise<BatchExecutionResult>;
    /**
     * Delete a document by id
     * @param id
     * @throws DocumentNotFoundError if the document does not exist
     * @returns void
     */
    delete(id: string): Promise<void>;
    /**
     * Delete a list of documents by id in batches.
     * It will send batches of 25 items synchronously.
     * @param ids document ids that will be deleted in batches
     * @returns {BatchExecutionResult}
     */
    batchDelete(ids: string[]): Promise<BatchExecutionResult>;
    /**
     * Find all documents in the table
     * @param limit
     * @param lastEvaluatedId the id of the last document in the previous page
     * @returns all documents in the table
     * example:
     * ```ts
     * let lastEvaluatedId: string | undefined = undefined
     * do {
     *   const users: User[] = await userRepository.findAll(10, lastEvaluatedId)
     *   lastEvaluatedId = users[users.length - 1]?.id || undefined
     * } while (lastEvaluatedId)
     * ```
     */
    findAll(limit?: number, lastEvaluatedId?: string): Promise<DocumentType[]>;
    /**
     * Find all documents in the table by query
     * @param query ScanCommandInput
     * @param lastEvaluatedId the id of the last document in the previous page
     * @returns all documents in the table
     * example:
     * ```ts
     * let lastEvaluatedId: string | undefined = undefined
     * do {
     *   const users: User[] = await userRepository.findAllBy(query, lastEvaluatedId)
     *   lastEvaluatedId = users[users.length - 1]?.id || undefined
     * } while (lastEvaluatedId)
     * ```
     */
    findAllBy(query: FindAllByQuery): Promise<DocumentType[]>;
    /**
     * Count all documents in the table
     * @returns the number of documents in the table
     */
    count(): Promise<number>;
    /**
     * Find by query
     * @param queryCommandInput
     * @returns all documents that match the query
     */
    queryBy(queryCommandInput: Omit<QueryCommandInput, 'TableName'>): Promise<DocumentType[]>;
    /**
     * Find one by query
     * @param queryCommandInput
     * @returns the first document that matches the query
     */
    queryOneBy(queryCommandInput: Omit<QueryCommandInput, 'TableName'>): Promise<DocumentType>;
    protected executeBatchRequests<RequestType extends WriteRequest>(queue: RequestType[]): Promise<InternalBatchExecutionResult<RequestType>>;
    is5xxError(err: DynamoDBServiceException): boolean;
    isRetryableError(err: unknown): boolean;
}
export {};
//# sourceMappingURL=SimpleDynamoRepository.d.ts.map