"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SimpleDynamoRepository = void 0;
const lib_dynamodb_1 = require("@aws-sdk/lib-dynamodb");
const SimpleRepositoryNotFoundError_1 = require("./SimpleRepositoryNotFoundError");
const client_dynamodb_1 = require("@aws-sdk/client-dynamodb");
// Dynamo is limited to batches of 25 items.
const MAX_BATCH_SIZE = 25;
const DEFAULT_MAX_RETRIES = 10;
// See https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Programming.Errors.html#Programming.Errors.MessagesAndCodes
const RETRYABLE_ERRORS = [
    'ProvisionedThroughputExceededException',
    'ThrottlingException',
    'ItemCollectionSizeLimitExceededException',
    'LimitExceededException',
    'ProvisionedThroughputExceeded',
    'RequestLimitExceeded',
    'UnrecognizedClientException',
];
const exponentialBackoff = (retries) => __awaiter(void 0, void 0, void 0, function* () {
    const ms = (2 ** retries - 1) * 700 + Math.random() * 600 + 50;
    yield new Promise((res) => setTimeout(res, ms));
});
/**
 * A simple repository for DynamoDB
 * @template DocumentType the type of the document
 * example:
 * ```ts
 * type User {
 *  id: string
 *  name: string
 * }
 * class UserRepository extends SimpleDynamoRepository<User> {}
 * const userRepository = new UserRepository(new DynamoDBClient({ region: 'eu-central-1' }), 'users')
 * ```
 */
class SimpleDynamoRepository {
    constructor(client, tableName, options) {
        var _a, _b, _c;
        this.client = client;
        this.tableName = tableName;
        this.backOffProvider = (_a = options === null || options === void 0 ? void 0 : options.backOffProvider) !== null && _a !== void 0 ? _a : exponentialBackoff;
        this.maxRetries = (_b = options === null || options === void 0 ? void 0 : options.maxRetries) !== null && _b !== void 0 ? _b : DEFAULT_MAX_RETRIES;
        this.batchSize = Math.min((_c = options === null || options === void 0 ? void 0 : options.batchSize) !== null && _c !== void 0 ? _c : MAX_BATCH_SIZE, MAX_BATCH_SIZE);
    }
    /**
     * Find a document by id
     * @param id
     * @throws DocumentNotFoundError if the document does not exist
     * @returns the document
     */
    findOne(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const doc = yield this.client.send(new lib_dynamodb_1.GetCommand({ Key: { id }, TableName: this.tableName }));
            if (!doc.Item)
                throw new SimpleRepositoryNotFoundError_1.SimpleRepositoryNotFoundError(id);
            return doc.Item;
        });
    }
    /**
     * Find a document by the global secondary index
     * @param key the name of the secondary index
     * @param value the value of the secondary index
     * @returns all documents with the given secondary index
     */
    findBy(secondaryKey, value) {
        return __awaiter(this, void 0, void 0, function* () {
            const queryCommandInput = {
                TableName: this.tableName,
                IndexName: secondaryKey,
                KeyConditionExpression: '#key = :value',
                ExpressionAttributeNames: { '#key': secondaryKey },
                ExpressionAttributeValues: { ':value': value },
            };
            const docs = yield this.client.send(new lib_dynamodb_1.QueryCommand(queryCommandInput));
            return docs.Items;
        });
    }
    /**
     * Find the first document by the global secondary index
     * @param key the name of the secondary index
     * @param value the value of the secondary index
     * @returns the first document with the given secondary index
     */
    findOneBy(secondaryKey, value) {
        return __awaiter(this, void 0, void 0, function* () {
            const queryCommandInput = {
                TableName: this.tableName,
                IndexName: secondaryKey,
                KeyConditionExpression: '#key = :value',
                ExpressionAttributeNames: { '#key': secondaryKey },
                ExpressionAttributeValues: { ':value': value },
            };
            const docs = yield this.client.send(new lib_dynamodb_1.QueryCommand(queryCommandInput));
            if (docs && docs.Items && docs.Items.length > 0)
                return docs.Items[0];
            throw new SimpleRepositoryNotFoundError_1.SimpleRepositoryNotFoundError();
        });
    }
    /**
     * Save a document
     * @param doc
     * @returns the saved document
     */
    save(doc) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.client.send(new lib_dynamodb_1.PutCommand({
                Item: doc,
                TableName: this.tableName,
            }));
            return doc;
        });
    }
    /**
     * Save a list of documents in batches.
     * It will send batches of 25 items synchronously.
     * @param documents list of documents that will be saved in batches
     * @returns {BatchExecutionResult}
     */
    batchSave(documents) {
        return __awaiter(this, void 0, void 0, function* () {
            const queue = documents.map((doc) => ({ PutRequest: { Item: doc } }));
            const result = yield this.executeBatchRequests(queue);
            const toDocumentId = (req) => req.PutRequest.Item.id;
            return { unprocessed: result.unprocessed.map(toDocumentId) };
        });
    }
    /**
     * Delete a document by id
     * @param id
     * @throws DocumentNotFoundError if the document does not exist
     * @returns void
     */
    delete(id) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.client.send(new lib_dynamodb_1.DeleteCommand({
                    Key: { id },
                    TableName: this.tableName,
                    ConditionExpression: 'attribute_exists(id)',
                }));
            }
            catch (error) {
                if (error.name === 'ConditionalCheckFailedException')
                    throw new SimpleRepositoryNotFoundError_1.SimpleRepositoryNotFoundError(id);
                else
                    throw error;
            }
        });
    }
    /**
     * Delete a list of documents by id in batches.
     * It will send batches of 25 items synchronously.
     * @param ids document ids that will be deleted in batches
     * @returns {BatchExecutionResult}
     */
    batchDelete(ids) {
        return __awaiter(this, void 0, void 0, function* () {
            const queue = ids.map((id) => ({ DeleteRequest: { Key: { id } } }));
            const result = yield this.executeBatchRequests(queue);
            const toDocumentId = (req) => req.DeleteRequest.Key.id;
            return { unprocessed: result.unprocessed.map(toDocumentId) };
        });
    }
    /**
     * Find all documents in the table
     * @param limit
     * @param lastEvaluatedId the id of the last document in the previous page
     * @returns all documents in the table
     * example:
     * ```ts
     * let lastEvaluatedId: string | undefined = undefined
     * do {
     *   const users: User[] = await userRepository.findAll(10, lastEvaluatedId)
     *   lastEvaluatedId = users[users.length - 1]?.id || undefined
     * } while (lastEvaluatedId)
     * ```
     */
    findAll(limit = 1000, lastEvaluatedId) {
        return __awaiter(this, void 0, void 0, function* () {
            const scanCommandInput = {
                TableName: this.tableName,
                Limit: limit,
            };
            if (lastEvaluatedId)
                scanCommandInput.ExclusiveStartKey = { id: lastEvaluatedId };
            const docs = yield this.client.send(new lib_dynamodb_1.ScanCommand(scanCommandInput));
            return docs.Items;
        });
    }
    /**
     * Find all documents in the table by query
     * @param query ScanCommandInput
     * @param lastEvaluatedId the id of the last document in the previous page
     * @returns all documents in the table
     * example:
     * ```ts
     * let lastEvaluatedId: string | undefined = undefined
     * do {
     *   const users: User[] = await userRepository.findAllBy(query, lastEvaluatedId)
     *   lastEvaluatedId = users[users.length - 1]?.id || undefined
     * } while (lastEvaluatedId)
     * ```
     */
    findAllBy(query) {
        return __awaiter(this, void 0, void 0, function* () {
            const scanCommandInput = Object.assign(Object.assign({}, query.scanCommandInput), { TableName: this.tableName });
            if (query === null || query === void 0 ? void 0 : query.lastEvaluatedId)
                scanCommandInput.ExclusiveStartKey = { id: query.lastEvaluatedId };
            const docs = yield this.client.send(new lib_dynamodb_1.ScanCommand(scanCommandInput));
            return docs.Items;
        });
    }
    /**
     * Count all documents in the table
     * @returns the number of documents in the table
     */
    count() {
        return __awaiter(this, void 0, void 0, function* () {
            const docs = yield this.client.send(new lib_dynamodb_1.ScanCommand({ TableName: this.tableName, Select: 'COUNT' }));
            return docs.Count || 0;
        });
    }
    /**
     * Find by query
     * @param queryCommandInput
     * @returns all documents that match the query
     */
    queryBy(queryCommandInput) {
        return __awaiter(this, void 0, void 0, function* () {
            const docs = yield this.client.send(new lib_dynamodb_1.QueryCommand(Object.assign(Object.assign({}, queryCommandInput), { TableName: this.tableName })));
            return docs.Items;
        });
    }
    /**
     * Find one by query
     * @param queryCommandInput
     * @returns the first document that matches the query
     */
    queryOneBy(queryCommandInput) {
        return __awaiter(this, void 0, void 0, function* () {
            const docs = yield this.client.send(new lib_dynamodb_1.QueryCommand(Object.assign(Object.assign({}, queryCommandInput), { TableName: this.tableName })));
            if (docs && docs.Items && docs.Items.length > 0)
                return docs.Items[0];
            throw new SimpleRepositoryNotFoundError_1.SimpleRepositoryNotFoundError();
        });
    }
    executeBatchRequests(queue) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            let retries = 0;
            while (queue.length > 0) {
                const batch = queue.splice(0, Math.min(queue.length, this.batchSize));
                const input = { RequestItems: { [this.tableName]: batch } };
                let failedRequests = [];
                try {
                    const result = yield this.client.send(new lib_dynamodb_1.BatchWriteCommand(input));
                    if (result.UnprocessedItems) {
                        failedRequests = (_a = result.UnprocessedItems) === null || _a === void 0 ? void 0 : _a[this.tableName];
                    }
                }
                catch (err) {
                    if (this.isRetryableError(err))
                        failedRequests = batch;
                    else
                        throw err;
                }
                queue.unshift(...failedRequests);
                if ((failedRequests === null || failedRequests === void 0 ? void 0 : failedRequests.length) > 0) {
                    if (retries >= this.maxRetries)
                        return { unprocessed: queue };
                    yield this.backOffProvider(retries);
                    retries++;
                }
            }
            return { unprocessed: queue };
        });
    }
    is5xxError(err) {
        var _a;
        const statusCode = (_a = err.$metadata) === null || _a === void 0 ? void 0 : _a.httpStatusCode;
        return statusCode != undefined && statusCode >= 500 && statusCode < 600;
    }
    isRetryableError(err) {
        return (err instanceof client_dynamodb_1.DynamoDBServiceException &&
            (this.is5xxError(err) || RETRYABLE_ERRORS.includes(err.name)));
    }
}
exports.SimpleDynamoRepository = SimpleDynamoRepository;
//# sourceMappingURL=SimpleDynamoRepository.js.map