"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SimpleS3Repository = void 0;
const client_s3_1 = require("@aws-sdk/client-s3");
const s3_request_presigner_1 = require("@aws-sdk/s3-request-presigner");
const SimpleRepositoryNotFoundError_1 = require("./SimpleRepositoryNotFoundError");
/**
 * A simple repository for S3 which is used to save and retrieve files and their metadata
 * @param T the type of the file
 * @param bucketName the name of the bucket
 * @param client the S3 client
 * @example
 * ```ts
 * const s3Client = new S3Client({ region: 'eu-central-1' })
 * const s3Repository = new SimpleS3Repository<SimpleFile>(s3Client, 'my-bucket')
 * ```
 */
class SimpleS3Repository {
    constructor(client, bucketName) {
        this.FIFTEEN_MINUTES = 15 * 60;
        this.client = client;
        this.bucketName = bucketName;
    }
    /**
     * Save a file
     * @param file as SimpleFile
     * @returns void
     * @example
     * ```ts
     * const file: SimpleFile = {
     *  key: 'foo/bar.json',
     *  content: 'my-file-content',
     *  contentType: 'application/json',
     * }
     * await s3Repository.save(file)
     * ```
     */
    save(file) {
        return __awaiter(this, void 0, void 0, function* () {
            const data = {
                Bucket: this.bucketName,
                Key: file.key,
                Body: file.content,
                ContentType: file.contentType,
                Meta: file,
            };
            yield this.client.send(new client_s3_1.PutObjectCommand(data));
        });
    }
    /**
     * Get a file as string
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the file as string
     */
    getOneAsString(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield (yield this.getOne(filePath)).transformToString();
        });
    }
    /**
     * Get a file as byte array
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the file as byte array
     */
    getOneAsByteArray(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.getOne(filePath)).transformToByteArray();
        });
    }
    /**
     * Get a file as readable stream
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the file as a readable stream
     */
    getOneAsReadableStream(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.getOne(filePath)).transformToWebStream();
        });
    }
    getOne(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.client.send(new client_s3_1.GetObjectCommand({ Bucket: this.bucketName, Key: filePath }));
            if (!response.Body)
                throw new SimpleRepositoryNotFoundError_1.SimpleRepositoryNotFoundError(filePath);
            return response.Body;
        });
    }
    /**
     * Get all which match the prefix
     * @param prefix the prefix
     * @returns all files which match the prefix
     */
    list(prefix) {
        return __awaiter(this, void 0, void 0, function* () {
            const results = [];
            let ContinuationToken = undefined;
            do {
                const input = {
                    Bucket: this.bucketName,
                    Prefix: prefix,
                    ContinuationToken,
                };
                const response = yield this.client.send(new client_s3_1.ListObjectsV2Command(input));
                if (response.Contents)
                    results.push(...response.Contents.map((object) => ({
                        filePath: object.Key,
                        lastModified: object.LastModified.toISOString(),
                    })));
                ContinuationToken = response.NextContinuationToken;
            } while (ContinuationToken);
            return results;
        });
    }
    /**
     * Get all metadata for a file
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the metadata of the file
     */
    getOneMetadata(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.client.send(new client_s3_1.HeadObjectCommand({ Bucket: this.bucketName, Key: filePath }));
            return response.Metadata;
        });
    }
    /**
     * Delete a file
     * @param filePath the path of the file
     * @returns void
     */
    deleteOne(filepath) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.delete([filepath]);
        });
    }
    /**
     * Delete multiple files
     * @param filePaths the paths of the files
     * @returns void
     */
    delete(filePaths) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!filePaths.length)
                return;
            const Objects = filePaths.map((filePath) => ({ Key: filePath }));
            yield this.client.send(new client_s3_1.DeleteObjectsCommand({ Bucket: this.bucketName, Delete: { Objects } }));
        });
    }
    /**
     * Get a signed download url
     * @param filePath the path of the file
     * @param expiresIn the time in seconds the url is valid
     * @returns the signed download url
     */
    getSignedDownloadUrl(filePath, expiresIn = this.FIFTEEN_MINUTES) {
        return __awaiter(this, void 0, void 0, function* () {
            const command = new client_s3_1.GetObjectCommand({
                Bucket: this.bucketName,
                Key: filePath,
            });
            return yield (0, s3_request_presigner_1.getSignedUrl)(this.client, command, { expiresIn });
        });
    }
    /**
     * Get a signed upload url
     * @param filePath the path of the file
     * @param expiresIn the time in seconds the url is valid
     * @returns the signed upload url
     */
    getSignedUploadUrl(filePath, expiresIn = this.FIFTEEN_MINUTES) {
        return __awaiter(this, void 0, void 0, function* () {
            const command = new client_s3_1.PutObjectCommand({
                Bucket: this.bucketName,
                Key: filePath,
            });
            return yield (0, s3_request_presigner_1.getSignedUrl)(this.client, command, { expiresIn });
        });
    }
}
exports.SimpleS3Repository = SimpleS3Repository;
//# sourceMappingURL=SimpleS3Repository.js.map