import {
  S3Client,
  ListObjectsV2Command,
  GetObjectCommand,
  DeleteObjectsCommand,
  PutObjectCommand,
  HeadObjectCommand,
  ListObjectsV2CommandInput,
} from '@aws-sdk/client-s3'
import { getSignedUrl } from '@aws-sdk/s3-request-presigner'
import { SimpleRepositoryNotFoundError } from './SimpleRepositoryNotFoundError'

export interface SimpleFileType {
  key: string
  content: Uint8Array | string | Buffer
  contentType: string
}

export interface SimpleRepositoryObject {
  filePath: string
  lastModified: string
}

/**
 * A simple repository for S3 which is used to save and retrieve files and their metadata
 * @param T the type of the file
 * @param bucketName the name of the bucket
 * @param client the S3 client
 * @example
 * ```ts
 * const s3Client = new S3Client({ region: 'eu-central-1' })
 * const s3Repository = new SimpleS3Repository<SimpleFile>(s3Client, 'my-bucket')
 * ```
 */
export class SimpleS3Repository<FileType extends SimpleFileType> {
  protected readonly client: S3Client
  protected readonly bucketName: string
  private readonly FIFTEEN_MINUTES = 15 * 60

  constructor(client: S3Client, bucketName: string) {
    this.client = client
    this.bucketName = bucketName
  }

  /**
   * Save a file
   * @param file as SimpleFile
   * @returns void
   * @example
   * ```ts
   * const file: SimpleFile = {
   *  key: 'foo/bar.json',
   *  content: 'my-file-content',
   *  contentType: 'application/json',
   * }
   * await s3Repository.save(file)
   * ```
   */
  async save(file: FileType): Promise<void> {
    const data = {
      Bucket: this.bucketName,
      Key: file.key,
      Body: file.content,
      ContentType: file.contentType,
      Meta: file,
    }

    await this.client.send(new PutObjectCommand(data))
  }

  /**
   * Get a file as string
   * @param filePath the path of the file
   * throws DocumentNotFoundError if the file does not exist
   * @returns the file as string
   */
  async getOneAsString(filePath: string): Promise<string> {
    return await (await this.getOne(filePath)).transformToString()
  }

  /**
   * Get a file as byte array
   * @param filePath the path of the file
   * throws DocumentNotFoundError if the file does not exist
   * @returns the file as byte array
   */
  async getOneAsByteArray(filePath: string): Promise<Uint8Array> {
    return (await this.getOne(filePath)).transformToByteArray()
  }

  /**
   * Get a file as readable stream
   * @param filePath the path of the file
   * throws DocumentNotFoundError if the file does not exist
   * @returns the file as a readable stream
   */
  async getOneAsReadableStream(filePath: string): Promise<ReadableStream> {
    return (await this.getOne(filePath)).transformToWebStream()
  }

  private async getOne(filePath: string) {
    const response = await this.client.send(
      new GetObjectCommand({ Bucket: this.bucketName, Key: filePath })
    )

    if (!response.Body) throw new SimpleRepositoryNotFoundError(filePath)

    return response.Body
  }

  /**
   * Get all which match the prefix
   * @param prefix the prefix
   * @returns all files which match the prefix
   */
  async list(prefix: string): Promise<SimpleRepositoryObject[]> {
    const results: SimpleRepositoryObject[] = []

    let ContinuationToken: string | undefined = undefined
    do {
      const input: ListObjectsV2CommandInput = {
        Bucket: this.bucketName,
        Prefix: prefix,
        ContinuationToken,
      }
      const response = await this.client.send(new ListObjectsV2Command(input))

      if (response.Contents)
        results.push(
          ...response.Contents.map((object) => ({
            filePath: object.Key as string,
            lastModified: (object.LastModified as Date).toISOString(),
          }))
        )

      ContinuationToken = response.NextContinuationToken
    } while (ContinuationToken)

    return results
  }

  /**
   * Get all metadata for a file
   * @param filePath the path of the file
   * throws DocumentNotFoundError if the file does not exist
   * @returns the metadata of the file
   */
  async getOneMetadata(filePath: string): Promise<FileType> {
    const response = await this.client.send(
      new HeadObjectCommand({ Bucket: this.bucketName, Key: filePath })
    )

    return response.Metadata as unknown as FileType
  }

  /**
   * Delete a file
   * @param filePath the path of the file
   * @returns void
   */
  async deleteOne(filepath: string): Promise<void> {
    await this.delete([filepath])
  }

  /**
   * Delete multiple files
   * @param filePaths the paths of the files
   * @returns void
   */
  async delete(filePaths: string[]): Promise<void> {
    if (!filePaths.length) return

    const Objects = filePaths.map((filePath) => ({ Key: filePath }))
    await this.client.send(
      new DeleteObjectsCommand({ Bucket: this.bucketName, Delete: { Objects } })
    )
  }

  /**
   * Get a signed download url
   * @param filePath the path of the file
   * @param expiresIn the time in seconds the url is valid
   * @returns the signed download url
   */
  async getSignedDownloadUrl(filePath: string, expiresIn = this.FIFTEEN_MINUTES): Promise<string> {
    const command = new GetObjectCommand({
      Bucket: this.bucketName,
      Key: filePath,
    })

    return await getSignedUrl(this.client, command, { expiresIn })
  }

  /**
   * Get a signed upload url
   * @param filePath the path of the file
   * @param expiresIn the time in seconds the url is valid
   * @returns the signed upload url
   */
  async getSignedUploadUrl(filePath: string, expiresIn = this.FIFTEEN_MINUTES): Promise<string> {
    const command = new PutObjectCommand({
      Bucket: this.bucketName,
      Key: filePath,
    })

    return await getSignedUrl(this.client, command, { expiresIn })
  }
}
